/*  game.c - VS Bomber: Game Engine
 *
 *  Copyright 1998, Benjamin C. W. Sittler and Stefan M. W. Sittler
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include "game.h"
#include "tse.h"

/* update handler */
volatile int update = 0;

void update_handler() {
   update ++;
}

void update_handler_end() { };

/* player table */
struct Player player[PLAYERS] = {
   {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
      {KEY_UP, KEY_DOWN, KEY_LEFT, KEY_RIGHT, KEY_5_PAD, KEY_ENTER, }, },
   {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
      {KEY_W, KEY_X, KEY_A, KEY_D, KEY_S, KEY_TAB, }, },
};

/* items to dispense */
int items = (ROWS * COLS) >> 2;

/* is poisoned dispensed? */
int poison = 0;

/* bomb table */
struct Bomb bomb[BOMBS] = { {0, }, };

/* construct composite characters */
/* bmp is a temporary working bitmap at least as large as a character */
void compose(BITMAP *bmp, int back, int src, int dest, int w, int h)
{
  int r, c;
  
  for (r = 0; r < h; r ++)
    for (c = 0; c < w; c ++) {
      blit(charset[back],
	   bmp,
	   0, 0, 0, 0, char_w, char_h);
      draw_sprite(bmp,
		  charset[CHARSET(r, src + c)],
		  0, 0);
      draw_sprite(charset[CHARSET(r, dest + c)],
		  bmp,
		  0, 0);
    }
}

/* is a cell blank? */
int isblank(int r, int c) {
  int ch = vs_screen[0][r][c] . ch;

  if ((ch == BG_CHAR) ||
      (ch == SHADOW_CHAR) ||
      (ch == WSHADOW_CHAR))
    return 1;
  return 0;
}

/* blanks a screen matrix cell (possibly adding shadows) */
void blank(int r, int c) {
  if ((r == 3) && (c > 2) && (c < 18)) {
    vs_screen[0][r][c] . ch = WSHADOW_CHAR;
    vs_screen[0][r][c] . lw = WSHADOW_LW;
    vs_screen[0][r][c] . delay = WSHADOW_D;
  } else if (r && (vs_screen[0][r - 1][c] . ch == HARD_CHAR)) {
    vs_screen[0][r][c] . ch = SHADOW_CHAR;
    vs_screen[0][r][c] . lw = SHADOW_LW;
    vs_screen[0][r][c] . delay = SHADOW_D;
  } else {
    vs_screen[0][r][c] . ch = BG_CHAR;
    vs_screen[0][r][c] . lw = BG_LW;
    vs_screen[0][r][c] . delay = BG_D;
  }
  vs_screen[0][r][c] . fr = 0;
  vs_screen[0][r][c] . st = 0;
  vs_screen[0][r][c] . timer = vs_screen[0][r][c] . delay;
  vs_screen[0][r][c] . dirty = 1;
  if (r) {
    if (vs_screen[1][r - 1][c] . ch)
      vs_screen[1][r - 1][c] . ch = 0;
  } else if (vs_screen[1][ROWS - 1][c] . ch)
    vs_screen[1][ROWS - 1][c] . ch = 0;
}

/* blasts a tile */
void blast(int r, int c) {
  int ch = vs_screen[0][r][c] . ch;

  if (ch == SOFT_CHAR) {
    vs_screen[0][r][c] . ch = MELT_CHAR;
    vs_screen[0][r][c] . lw = MELT_LW;
    vs_screen[0][r][c] . fr = 0;
    vs_screen[0][r][c] . st = 0;
    vs_screen[0][r][c] . delay = MELT_D;
    vs_screen[0][r][c] . timer = vs_screen[0][r][c] . delay;
    vs_screen[0][r][c] . dirty = 1;
  } else if (ch == ITEM_CHAR) {
    if (vs_screen[0][r][c] . st == ITEM_ST_POISON)
       poison = 0;
    vs_screen[0][r][c] . ch = NUKE_CHAR;
    vs_screen[0][r][c] . lw = NUKE_LW;
    vs_screen[0][r][c] . fr = 0;
    vs_screen[0][r][c] . st = 0;
    vs_screen[0][r][c] . delay = NUKE_D;
    vs_screen[0][r][c] . timer = vs_screen[0][r][c] . delay;
    vs_screen[0][r][c] . dirty = 1;
  }
}

/* mixes explosion stances */
void boom_mix(int r, int c, int st) {
  int st1;

  st1 = vs_screen[0][r][c] . st;
  if ((vs_screen[0][r][c] . ch != BOOM_CHAR) ||
      vs_screen[0][r][c] . fr ||
      (st == BOOM_ST_CENTER) ||
      (st1 == st))
    st1 = st;
  else if ((st1 == BOOM_ST_TOP) ||
	   (st1 == BOOM_ST_VERT) ||
	   (st1 == BOOM_ST_BOTTOM)) {
    if ((st == BOOM_ST_TOP) ||
	(st == BOOM_ST_VERT) ||
	(st == BOOM_ST_BOTTOM))
      st1 = BOOM_ST_VERT;
    else
      st1 = BOOM_ST_CENTER;
  } else if ((st1 == BOOM_ST_LEFT) ||
	     (st1 == BOOM_ST_HORIZ) ||
	     (st1 == BOOM_ST_RIGHT)) {
    if ((st == BOOM_ST_LEFT) ||
	(st == BOOM_ST_HORIZ) ||
	(st == BOOM_ST_RIGHT))
      st1 = BOOM_ST_HORIZ;
    else
      st1 = BOOM_ST_CENTER;
  } else
    st1 = BOOM_ST_CENTER;
  vs_screen[0][r][c] . st = st1;
  vs_screen[0][r][c] . ch = BOOM_CHAR;
  vs_screen[0][r][c] . lw = BOOM_LW;
  vs_screen[0][r][c] . fr = 0;
  vs_screen[0][r][c] . delay = BOOM_D;
  vs_screen[0][r][c] . timer = vs_screen[0][r][c] . delay;
  vs_screen[0][r][c] . dirty = 1;
}

/* initialize a player */
void init_player(int p)
{
   int s;

   player[p] . active = 0;
   player[p] . bombs = 1;
   player[p] . range = 2;
   player[p] . alive = 1;
   player[p] . kick = 0;
   player[p] . remote = 0;
   player[p] . poison = 0;
   player[p] . poisoned = 0;
   player[p] . vector = 0;
   player[p] . time = 0;
   player[p] . delay = 0;
   player[p] . throw = 0;
   s = MAN_S(p);
   sprite[s] . used = 1;
   sprite[s] . r = 3 + p * 10;
   sprite[s] . c = 4 + p * 12;
   sprite[s] . x = 0;
   sprite[s] . y = (MAN_H * char_h) / -2;
   sprite[s] . z = - (char_h * (1 + sprite[s] . r) + sprite[s] . y);
   sprite[s] . w = MAN_W;
   sprite[s] . h = MAN_H;
   sprite[s] . ch = MAN_CHAR + p * MAN_W * MAN_LW;
   sprite[s] . fr = 1;
   sprite[s] . st = MAN_ST_DOWN;
   sprite[s] . mst = sprite[s] . st;
   sprite[s] . lw = MAN_LW;
   sprite[s] . timer = 0;
   sprite[s] . delay = 0;
   dirty_sprite(s);
}

/* tests whether a screen matrix cell may be kicked onto */
int kickable(int b, int r, int c) {
  int ch = vs_screen[0][r][c] . ch, p, b1, s;
  
  for (b1 = 0; b1 < BOMBS; b1 ++)
    if (sprite[(s = BOMB_S(b1))] . used &&
	(sprite[s] . r == r) &&
	(sprite[s] . c == c))
      return 0;
  for (p = 0; p < PLAYERS; p ++)
    if (sprite[(s = MAN_S(p))] . used &&
	(sprite[s] . r == r) &&
	(sprite[s] . c == c))
      return 0;
  if ((ch == BOOM_CHAR) ||
      isblank(r, c))
    return 1;
  return 0;
}

/* tests whether a screen matrix cell may be walked onto */
int walkable(int p, int dx, int dy) {
   int ch, b, s, r, c;

   r = sprite[(s = MAN_S(p))] . r + dy;
   c = sprite[s] . c + dx;
   ch = vs_screen[0][r][c] . ch;
   for (b = 0; b < BOMBS; b ++)
     if (sprite[(s = BOMB_S(b))] . used &&
	 (sprite[s] . r == r) &&
	 (sprite[s] . c == c)) {
       if (player[p] . kick && kickable(b, r + dy, c + dx)) {
	 bomb[b] . dx = dx;
	 bomb[b] . dy = dy;
       } else
	 return 0;
     }
   if ((ch == BOOM_CHAR) ||
       (ch == ITEM_CHAR) ||
       isblank(r, c))
      return 1;
   return 0;
}

int autoskip = 0;
int frameskip = -1, vframe = 0, frame = 0;

/* update and wait for refresh */
/* BMP is a temporary working tile at least as large as a character */
void refresh(BITMAP *bmp, int unthrottled, int upper) {
  vframe ++;
  if (! (vframe % RIPPLE_D)) {
    ripple ++;
    ripple %= char_h;
    ripple_h = ripple / 2 + 1;
  }
  if (! (frameskip && (vframe % frameskip))) {
    if (! unthrottled) {
      /* try to run at a constant speed */
      if (! update) {
        if (autoskip && frameskip)
           frameskip --;
        while (! update);
      } else if (autoskip) {
        if ((frameskip < MAXSKIP) &&
            (update > frameskip))
           frameskip ++;
        else if  (frameskip &&
                  (update < frameskip))
           frameskip --;
      }
      update = 0;
    }
    frame ++;
    update_screen(bmp, upper);
  }
}
