/*  game.h - VS Bomber: Game Engine Interface
 *
 *  Copyright 1998, Benjamin C. W. Sittler and Stefan M. W. Sittler
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef game_h_inc
#define game_h_inc

#include <allegro.h>

/* update handler */
extern volatile int update;

void update_handler();
void update_handler_end();

/* players */
#define PLAYERS 2

/* simulataneous bombs */
#define BOMBS (PLAYERS * 13)

/* sprite table size */
#define SPRITES (PLAYERS + BOMBS + 2 + BOMBS + 1)

/* sprite index macros */
#define MAN_S(m)        (m)
#define BOMB_S(b)       MAN_S(PLAYERS + (b))
#define BSHADOW_S(b)    BOMB_S(BOMBS + (b))
#define PAUSE_S         BSHADOW_S(BOMBS)
#define OVER_S          (PAUSE_S + 1)
#define POISON_S        (OVER_S + 1)

/* maximum automatic frameskip */
#define MAXSKIP 20

/* player table */
struct Player {
   int active; /* active bombs */
   int bombs; /* possible simultaneous bombs */
   int range; /* explosion range */
   int alive; /* is player alive? */
   int kick; /* can player kick? */
   int remote; /* how many of the player's bombs are remote-controlled? */
   int poisoned; /* how long is player poisoned? */
   int poison; /* how is player poisoned? */
   int vector; /* does the player have poison vector immunity? */
   int time; /* how many of the player's bombs are time bombs? */
   int delay; /* how long has the player been winding the clock? */
   int throw; /* can player throw? */
   int keys[6]; /* movement keys - u d l r bomb special */
   int special; /* handled special keypress */
};

extern struct Player player[PLAYERS];

/* items to dispense */
extern int items;

/* is poisoned dispensed? */
extern int poison;

/* bomb table */
struct Bomb {
   int fuse; /* explosion timer */
   int p; /* owner */
   int dx, dy; /* velocity */
};

extern struct Bomb bomb[BOMBS];

/* ripple delay */
#define RIPPLE_D        ((char_h > 3) ? (char_h / 4) : 1)

/* basic map tiles */
#define BG_CHAR         CHARSET(0,1)
#define BG_LW           1
#define BG_D            0

#define HARD_CHAR       CHARSET(0,2)
#define HARD_LW         1
#define HARD_D          0

#define SHADOW_CHAR     CHARSET(1,2)
#define SHADOW_LW       1
#define SHADOW_D        0

#define WSHADOW_CHAR    CHARSET(13,14)
#define WSHADOW_LW      1
#define WSHADOW_D       0

#define SOFT_CHAR       CHARSET(0,3)
#define SOFT_LW         1
#define SOFT_D          0

#define MELT_CHAR       CHARSET(0,4)
#define MELT_LW         4
#define MELT_D          7

#define NUKE_CHAR       CHARSET(1,10)
#define NUKE_LW         7
#define NUKE_D          4

/* item tiles */
#define ITEMS 7
#define ITEM_CHAR       CHARSET(0,9)
#define ITEM_BG_CHAR    CHARSET(0,13)
#define ITEM_LW         1
#define ITEM_ST_FLAME   0
#define ITEM_ST_BOMB    1
#define ITEM_ST_KICK    2
#define ITEM_ST_REMOTE  3
#define ITEM_ST_POISON  4
#define ITEM_ST_TIME    5
#define ITEM_ST_THROW   6
#define ITEM_D          0

/* explosion tiles */
#define BOOM_CHAR       CHARSET(11,0)
#define BOOM_LW         3
#define BOOM_ST_TOP     0
#define BOOM_ST_VERT    1
#define BOOM_ST_BOTTOM  2
#define BOOM_ST_LEFT    3
#define BOOM_ST_HORIZ   4
#define BOOM_ST_RIGHT   5
#define BOOM_ST_CENTER  6
#define BOOM_D          7

/* border tiles */
#define LWALL_CHAR      CHARSET(11, 13)
#define LWALL_H         7
#define LWALL_LW        1
#define LWALL_ST_0      2
#define LWALL_LH        3
#define LWALL_D         0

#define RWALL_CHAR      CHARSET(11, 17)
#define RWALL_H         7
#define RWALL_LW        1
#define RWALL_ST_0      2
#define RWALL_LH        3
#define RWALL_D         0

#define HWALL_CHAR      CHARSET(11, 14)
#define HWALL_CH1       CHARSET(16, 14)
#define HWALL_W         3
#define HWALL_H         2
#define HWALL_LW        1
#define HWALL_D         0

/* player sprites */
#define MAN_CHAR        CHARSET(2,3)
#define MAN_W           1
#define MAN_H           2
#define MAN_D           10
#define MAN_DX_SLOW     ((char_w > 9) ? ((char_w + 10) / 20) : 1)
#define MAN_DY_SLOW     ((char_h > 9) ? ((char_h + 10) / 20) : 1)
#define MAN_DX          ((char_w > 4) ? ((char_w + 5) / 10) : 1)
#define MAN_DY          ((char_h > 4) ? ((char_h + 5) / 10) : 1)
#define MAN_DX_FAST     ((char_w > 2) ? ((char_w + 2) / 5) : 1)
#define MAN_DY_FAST     ((char_h > 2) ? ((char_h + 2) / 5) : 1)
#define MAN_LW          3
#define MAN_ST_CLICK    0
#define MAN_ST_DOWN     1
#define MAN_ST_UP       2
#define MAN_ST_LEFT     3
#define MAN_ST_RIGHT    4

/* walking per-frame step size */
#define X_STEP          6
#define Y_STEP          6

/* death sprites */
#define DEATH_CHAR      CHARSET(12,3)
#define DEATH_W         1
#define DEATH_H         2
#define DEATH_LW        10
#define DEATH_D         15

/* bomb sprites */
#define BOMB_CHAR       CHARSET(2,0)
#define BOMB_W          1
#define BOMB_H          1
#define BOMB_DX         ((char_w > 2) ? ((char_w + 3) / 6) : 1)
#define BOMB_DY         ((char_h > 2) ? ((char_h + 3) / 6) : 1)
#define BOMB_LW         3
#define BOMB_D          10
#define BOMB_ST_NORM    0
#define BOMB_ST_REMOTE  1
#define BOMB_ST_SMALL   2
#define BOMB_ST_TIME    3
#define BOMB_ST_TIME_H  5
#define BOMB_FUSE       100

/* bomb shadow sprites */
#define BSHADOW_CHAR    CHARSET(10,0)
#define BSHADOW_W       1
#define BSHADOW_H       1
#define BSHADOW_LW      3
#define BSHADOW_D       10

/* chain-reaction fuse length */
#define CHAIN_FUSE      5

/* pause sprite */
#define PAUSE_CHAR      CHARSET(5,10)
#define PAUSE_W         7
#define PAUSE_H         6
#define PAUSE_DY        ((char_h > 1) ? ((char_h + 3) / 5) : 1)
#define PAUSE_LW        1
#define PAUSE_D         0

/* pause overlay */
#define OVER_R          1
#define OVER_C          1
#define OVER_CHAR       CHARSET(5,16)
#define OVER_W          1
#define OVER_H          1
#define OVER_LW         3
#define OVER_D          2

/* poison marker */
#define POISON_CHAR     CHARSET(4,10)
#define POISON_W        1
#define POISON_H        1
#define POISON_LW       1
#define POISON_D        10

/* poison types */
#define POISONS         5
#define POISON_SMALL    0
#define POISON_BOMB0    1
#define POISON_BOMB1    2
#define POISON_SLOW     3
#define POISON_FAST     4
#define POISON_T        1500

/* initial pause overlay delay */
#define BLINK_D         100

/* construct composite characters */
/* bmp is a temporary working bitmap at least as large as a character */
void compose(BITMAP *bmp, int back, int src, int dest, int w, int h);

/* is a cell blank? */
int isblank(int r, int c);

/* blanks a screen matrix cell (possibly adding shadows) */
void blank(int r, int c);

/* blasts a tile */
void blast(int r, int c);

/* mixes explosion stances */
void boom_mix(int r, int c, int st);

/* initialize a player */
void init_player(int p);

/* tests whether a screen matrix cell may be kicked onto */
int kickable(int b, int r, int c);

/* tests whether a screen matrix cell may be walked onto */
int walkable(int p, int dx, int dy);

extern int autoskip;
extern int frameskip, vframe, frame;

/* update and wait for refresh */
/* BMP is a temporary working tile at least as large as a character */
void refresh(BITMAP *bmp, int unthrottled, int upper);

#endif /* ! defined(game_h_inc) */

