/*  tse.h - VS Bomber: Tile and Sprite Engine Interface
 *
 *  Copyright 1998, Benjamin C. W. Sittler and Stefan M. W. Sittler
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef tse_h_inc
#define tse_h_inc

#include "game.h"

/* character table size */
#define CHARSET_COLS 20
#define CHARSET_ROWS 18

/* optional constants for performance optimization */
/* #define char_w 32 */
/* #define char_h 32 */

/* video mode */
extern int screen_w, screen_h, screen_depth;

/* character size */
#ifndef char_w
extern int char_w;
#endif
#ifndef char_h
extern int char_h;
#endif

/* renderers */
#define TSE_STYLES  4
#define TSE_GOOD    0
#define TSE_FAST    1
#define TSE_VSYNC   2
#define TSE_RIPPLE  3
#define TSE_NAME(t) \
(((t) == TSE_GOOD) ? "perfect renderer" : \
 (((t) == TSE_FAST) ? "fast renderer" : \
  (((t) == TSE_VSYNC) ? "fast renderer w/vsync" : \
   (((t) == TSE_RIPPLE) ? "ripple renderer" : \
    "unknown renderer"))))

extern int tse_style;

/* ripple scanline in characters */
extern int ripple, ripple_h;

/* screen matrix size */
#define COLS 21
#define ROWS 15

/* character set index macro */
#define CHARSET(row, col) ((row) * CHARSET_COLS + (col))

/* character set (sub-)bitmaps */
extern BITMAP *charset[CHARSET_COLS * CHARSET_ROWS];

/* screen matrices */
#define VS_SCREENS 2

extern struct Cell {
   int ch; /* charset coordinates */
   int dirty; /* dirty bit */
   int fr, st; /* frame and stance offsets */
   int lw; /* frame loop size */
   int timer, delay; /* frame advance timer */
} vs_screen[VS_SCREENS][ROWS][COLS];

/* sprite table */
extern struct Sprite {
   int used; /* sprite active? */
   int r, c; /* coordinates in screen matrix */
   int x, y; /* pixel offset from screen matrix coordinates */
   int w, h; /* size in tiles */
   int ch; /* charset coordinates */
   int fr, st; /* frame and stance offsets */
   int mst; /* movement-induced stance offset */
   int lw; /* frame loop size */
   int z, prio; /* depth, priority */
   int timer, delay; /* frame advance timer */
} sprite[SPRITES];

/* sprite priority table */
extern int sprite_prio[SPRITES];

/* pixel offsets of screen matrix from viewport's upper left corner */
extern int x_offset, y_offset;

/* dirty the region occupied by a sprite, re-sort it in the priority table */
extern void dirty_sprite(int s);

/* updates a sprite's frame timer, returns 1 when the sprite loops */
/* if dirty is NULL, dirty_sprite(s) is called; otherwise *dirty is set */
/* A B C B, A B C B, ... */
extern int sprite_loop(int s, int *dirty);

/* updates a sprite's frame timer, returns 1 when the sprite repeats */
/* if dirty is NULL, dirty_sprite(s) is called; otherwise *dirty is set */
/* A B C, A B C, ... */
extern int sprite_repeat(int s, int *dirty);

/* update screen from screen matrices and sprite table */
extern void update_screen(BITMAP *bmp, int upper);

#endif /* ! defined(tse_h_inc) */
