/*  vs.c - VS Bomber: Main Program
 *
 *  Copyright 1998, Benjamin C. W. Sittler and Stefan M. W. Sittler
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#define SUMMARY \
   "Summary: %s [-h] [-V] [options]\n"

#define USAGE \
"Usage: %s [options]\n" \
"Options:\n" \
"(many options can be set in [vs] section of Allegro CFG file)\n" \
"-f NUM          skip NUM frames for every one displayed (frameskip=%d)\n" \
"                -1 indicates automatic adjustment\n" \
"-g NUM          use graphics driver NUM (gfx_card=%d)\n" \
"                see list in Allegro CFG file\n" \
"-h              show this message and exit\n" \
"-L              print logo, and exit\n" \
"-o FILE         write the graphics characters to FILE after color-conversion\n" \
"-r W[xH[xD]]    run in WxHxD-bit graphics mode (gfx_res=%dx%dx%d-bit)\n" \
"-R NUM          use tile and sprite engine NUM (tse_style=%d)\n" \
"                see list below\n" \
"-s W[xH]        use WxH-pixel graphics characters (char_size=%dx%d-pixel)\n" \
"-S NUM          update every NUM ms (speed=%d)\n" \
"-t FILE         use the graphics characters from FILE (tiles=%s)\n" \
"-u              run in \"unthrottled\" mode (don't wait for a timer)\n" \
"-V              print version and disclaimer of warranty, and exit\n"

#define VERSION \
   "VS Bomber, release 0.3\n" \
   "Copyright 1998, Benjamin C. W. Sittler and Stefan M. W. Sittler\n" \
   "\n" \
   "This program is free software; you can redistribute it and/or modify\n" \
   "it under the terms of the GNU General Public License as published by\n" \
   "the Free Software Foundation; either version 2 of the License, or\n" \
   "(at your option) any later version.\n" \
   "\n" \
   "This program is distributed in the hope that it will be useful,\n" \
   "but WITHOUT ANY WARRANTY; without even the implied warranty of\n" \
   "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n" \
   "GNU General Public License for more details.\n" \
   "\n" \
   "You should have received a copy of the GNU General Public License\n" \
   "along with this program; if not, write to the Free Software\n" \
   "Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.\n" \
   "\n" \
   "Built with %s\n" \

#define LOGO \
"+  . : . ' ` .#  , ' ` , .$  ,\"  ,#  , . \n" \
",  '\"  :\"  ;\"  ` '   , '   : ;!  :#  '!  ` ' \n" \
"+  ; :   . '\"  ;%  ;\"  :#  :\"  `!. \n" \
"*  ;   ;!  ` .   :%  :\"  ; :#  :%  ` . \n" \
")  ;   :#  :   :%  `!. '   :#  :   , .#  '&  , : ' \n" \
"* !`\". '\"  `\". '&  `!. '!  `#. '$  , : ' \n" \
"M  , : ' \nJ  , : ' \nG  , : ' \nD  , : ' \nA  , : ' \n" \
">  , : '' $.$  :\"  : \n;  , : ')  :#  `#  ; :!  ; : \n" \
"8  , : ',  `#.#  :   :   ;   : \n5  , : '4  `\"  ;   : ;!  : \n" \
"2  , : '2  :#. '!  ,!  ;\"  : . \n" \
"/  , : '=  ; \n,  , : '7  ` :\"  :!  :\" $. \n" \
")  , : ';  :!  ; :!  '!  :#  ` \nH  :   ;   :   :\"  `#. \n" \
"H  : ;!  : ;(  ` \nH  :\"  :$  :#. ' \n"

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include "game.h"
#include "tse.h"

/* parse graphics resolution string of the form W[xH[xD]]
   assumes 4:3 for W:H ratio if H is not given */
int gfx_res(char *res) {
   if ((! *res) ||
       (! (screen_w = (int)strtol(res, &res, 0)))) {
       fprintf(stderr, "found [%s] instead of non-zero screen width\n",
               res);
       return 1;
   }
   if (! *res) {
      screen_h = 300000 * screen_w / 400000;
      return 0;
   }
   if ((*res != 'x') && (*res != 'X')) {
      fprintf(stderr, "found '%c' instead of 'x' after screen width\n",
                       *res);
      return 1;
   }
   if ((! *res) ||
       (! (screen_h = (int)strtol(res + 1, &res, 0)))) {
      fprintf(stderr, "found [%s] instead of non-zero screen height\n",
              res);
      return 1;
   }
   if (! *res)
      return 0;
   if ((*res != 'x') && (*res != 'X')) {
      fprintf(stderr, "found '%c' instead of 'x' after screen height\n",
                       *res);
      return 1;
   }
   if ((! *res) ||
       (! (screen_depth = (int)strtol(res + 1, &res, 0)))) {
      fprintf(stderr, "found [%s] instead of non-zero screen depth\n",
              res);
      return 2;
   } else if (*res) {
      fprintf(stderr, "ignoring garbage [%s] after screen depth\n",
              res);
   }
   return 0;
}


/* parse character size string of the form W[xH]
   assumes 1:1 for W:H ratio if H is not given */
int char_size(char *size) {
  int w, h;
  
  if ((! *size) ||
      (! (w = (int)strtol(size, &size, 0)))) {
    fprintf(stderr, "found [%s] instead of non-zero character width\n",
	    size);
    return 1;
  } else {
#ifndef char_w
    char_w = w;
#else
    if (w != char_w) {
      fprintf(stderr,
	      "char_w was fixed at %d pixels at compile-time, you need to recompile with variable char_w or constant char_w of %d\n",
	      char_w, w);
      return 1;
    }
#endif
  }
  if (! *size) {
#ifndef char_h
    char_h = char_w;
    return 0;
#else
    if (char_w != char_h) {
      fprintf(stderr,
	      "char_h was fixed at %d pixels at compile-time, you need to recompile with variable char_h or constant char_h of %d\n",
	      char_h, char_w);
      return 1;
    }
#endif
  }
  if ((*size != 'x') && (*size != 'X')) {
    fprintf(stderr, "found '%c' instead of 'x' after character width\n",
	    *size);
    return 1;
  }
  if ((! *size) ||
      (! (h = (int)strtol(size + 1, &size, 0)))) {
    fprintf(stderr, "found [%s] instead of non-zero character height\n",
	    size);
    return 1;
  } else if (*size) {
    fprintf(stderr, "ignoring garbage [%s] after character height\n",
	    size);
  }
#ifndef char_h
  char_h = h;
  return 0;
#else
  if (h != char_h) {
    fprintf(stderr,
	    "char_h was fixed at %d pixels at compile-time, you need to recompile with variable char_h or constant char_h of %d\n",
	    char_h, h);
    return 1;
  }
#endif
  return 0;
}

/* print an RLE ascii-art picture */
void rlescii(char *pic) {
   int n, c;

   while ((n = *pic ++) && (c = *pic ++))
     while ((n --) > 31)
       putchar(c);
}

int main(int argc, char **argv)
{
  char *tiles = "alltiles.bmp", *optiles = 0;
  char *shot = "shot.bmp";
  BITMAP *bmp;
  RGB pal[256];
  int upper = 1;
  int transparent;
  int unthrottled = 0;
  int gfx_card = GFX_AUTODETECT;
  int x, y, p, b, s, opt;
  int speed = 20;
  int redraw = 0;
  char *res = 0, *size = 0;

  if (allegro_init()) {
     fprintf(stderr, "%s: allegro_init failed\n", *argv);
     return 1;
  }
  if ((res = get_config_string("vs", "gfx_res", 0)) && gfx_res(res))
     return 2;
  else
     res = 0;
  if ((size = get_config_string("vs", "char_size", 0)) && char_size(size))
     return 2;
  else
     size = 0;
  frameskip = get_config_int("vs", "frameskip", frameskip);
  gfx_card = get_config_int("vs", "gfx_card", gfx_card);
  tse_style = get_config_int("vs", "tse_style", tse_style);
  speed = get_config_int("vs", "speed", speed);
  tiles = get_config_string("vs", "tiles", tiles);
  while ((opt = getopt(argc, argv, "f:g:hLo:r:R:s:S:t:uV")) != -1) {
    switch (opt) {
    case 'f':
      frameskip = (int)strtol(optarg, 0, 0);
      break;
    case 'g':
      gfx_card = (int)strtol(optarg, 0, 0);
      break;
    case 'h':
      printf(USAGE,
	     *argv,
	     frameskip,
	     gfx_card,
	     screen_w, screen_h, screen_depth,
	     tse_style,
	     char_w, char_h,
	     speed,
	     tiles);
      printf("\nSupported tile and sprite engines:\n");
      for (s = 0; s < TSE_STYLES; s ++)
	printf("%d\t %s%s\n", s, TSE_NAME(s),
	       (s == tse_style) ? " (default)" : "");
      return 0;
      break;
    case 'L':
      rlescii(LOGO);
      return 0;
      break;
    case 'o':
      optiles = optarg;
      break;
    case 'r':
      res = optarg;
      break;
    case 'R':
      tse_style = (int)strtol(optarg, 0, 0);
      break;
    case 's':
      size = optarg;
      break;
    case 'S':
      speed = (int)strtol(optarg, 0, 0);
      break;
    case 't':
      tiles = optarg;
      break;
    case 'u':
      unthrottled = 1;
      break;
    case 'V':
      printf(VERSION, allegro_id);
      return 0;
      break;
    default:
      fprintf(stderr, SUMMARY, *argv);
      return 2;
    }
  }
  if (frameskip < 0) {
    frameskip = 0;
    autoskip = 1;
  }
  if (res && gfx_res(res))
     return 2;
  if (size && char_size(size))
     return 2;
  if (tse_style >= TSE_STYLES) {
    fprintf(stderr, "%s: %d is not a valid tile and sprite engine\n",
	    *argv, tse_style);
    return 2;
  }
  if (optind < argc) {
     fprintf(stderr, "%s: no non-option arguments are permitted\n", *argv);
     return 2;
  }
  y_offset = (screen_h - ROWS * char_h) / 2;
  x_offset = (screen_w - COLS * char_w) / 2;
  if ((ROWS - 4 > screen_h / char_h) ||
      (COLS - 2 > screen_w / char_w)) {
     fprintf(stderr,
             "%dx%d screen matrix with %dx%d-pixel tiles doesn't fit in a %dx%d-pixel gfx. mode; %dx%d-pixel or larger needed; %dx%d-pixel preferred\n",
             COLS - 2, ROWS - 4, char_w, char_h, screen_w, screen_h,
	     (COLS - 2) * char_w,
	     (ROWS - 4) * char_h,
	     (COLS - 1) * char_w,
	     ROWS * char_h);
     return 2;
  } else if (ROWS - 2 > screen_h / char_h)
     y_offset -= char_h;
  set_color_depth(screen_depth);
  printf(VERSION, allegro_id);
  srand((int)time((time_t *) 0));
  printf("Timer: ");
  if (install_timer()) {
     printf("failed\n");
     fprintf(stderr, "%s: install_timer failed\n", *argv);
     return 1;
  }
  printf("ok\n");
  printf("Updates: ");
  if (! unthrottled) {
     LOCK_VARIABLE(update);
     LOCK_FUNCTION(update_handler);
     if (install_int(update_handler, speed)) {
        printf("%dms timer request failed\n", speed);
        fprintf(stderr, "%s: install_int failed\n", *argv);
        return 1;
     }
     printf("every %dms\n", speed);
  } else {
     printf("unthrottled\n");
  }
  printf("Keyboard: ");
  if (install_keyboard()) {
     printf("failed\n");
     fprintf(stderr, "%s: install_keyboard failed\n", *argv);
     return 1;
  }
  printf("ok\n");
  printf("%dx%d-pixel tiles from [%s]: *\b", char_w, char_h, tiles);
  fflush(stdout);
  /* load tileset */
  {
    if (! (bmp = load_bitmap(tiles, pal))) {
      printf("failed\n");
      fprintf(stderr, "%s: load_bitmap failed for %s\n", *argv, tiles);
      return 1;
    }
    if (optiles) {
      if (save_bitmap(optiles, bmp, pal)) {
	printf("failed\n");
	fprintf(stderr, "%s: save_bitmap failed for %s\n",
		*argv, optiles);
	return 1;
      }
      printf("saved to [%s] ", optiles);
    } else {
      printf("loaded ");
    }
    fflush(stdout);
  }
  putchar('@');
  putchar('\b');
  fflush(stdout);
  /* check to make sure it's big enough */
  if ((CHARSET_ROWS * char_h > bmp -> h) ||
      (CHARSET_COLS * char_w > bmp -> w)) {
     printf("failed");
     fprintf(stderr,
             "%s: [%s] is %dx%d pixels; %dx%d are needed for a %dx%d char. set\n",
             *argv,
             tiles,
             bmp -> w, bmp -> h,
             CHARSET_COLS * char_w, CHARSET_ROWS * char_h,
             CHARSET_COLS, CHARSET_ROWS);
     return 2;
  }
  /* move transparent pixels to proper color */
  {
     int trans = 0;

     transparent = getpixel(bmp, 0, 0);
     if (screen_depth <= 8) {
        RGB black = {0, 0, 0};

        pal[transparent] = pal[trans];
        pal[trans] = black;
     } else
        trans = makecol(255, 0, 255);
     if (transparent) {
        for (y = 0; y < bmp -> h; y ++) {
           for (x = 0; x < bmp -> w; x ++) {
              int col = getpixel(bmp, x, y);

              if (col == transparent)
                 putpixel(bmp, x, y, trans);
              else if (col == trans)
                 putpixel(bmp, x, y, transparent);
           }
           putchar(("/-\\|")[y & 3]);
           putchar('\b');
           fflush(stdout);
        }
     }
  }
  /* initialize charset */
  {
    int r, c;
    BITMAP *tmp;

    for (r = 0, y = 0; r < CHARSET_ROWS; r ++, y += char_h) {
      for (c = 0, x = 0; c < CHARSET_COLS; c ++, x += char_w) {
	if (! (tmp =
	       (charset[CHARSET(r, c)] =
		create_bitmap(char_w, char_h)))) {
	  printf("failed\n");
	  fprintf(stderr,
		  "%s: create_bitmap (%dx%d) failed for tile at col %d of row %d\n",
		  *argv, char_w, char_h, c, r);
	  return 1;
        } else {
	  blit(bmp,
	       tmp,
	       x, y,
	       0, 0,
	       char_w, char_h);
	}
      }
      putchar(("/-\\|")[r & 3]);
      putchar('\b');
      fflush(stdout);
    }
  }
  destroy_bitmap(bmp);
  fflush(stdout);
  /* create working bitmap */
  if (! (bmp = create_bitmap(char_w * 2, char_h * 2))) {
    printf("failed\n");
    fprintf(stderr, "%s: create_bitmap (%dx%d) failed\n",
	    *argv, char_w * 2, char_h * 2);
    return 1;
  }
  /* composite blocks, items, and explosions w/b.g. tiles */
  compose(bmp, BG_CHAR, HARD_CHAR, HARD_CHAR, HARD_LW, 1);
  compose(bmp, BG_CHAR, SHADOW_CHAR, SHADOW_CHAR, SHADOW_LW, 1);
  compose(bmp, BG_CHAR, WSHADOW_CHAR, WSHADOW_CHAR, WSHADOW_LW, 1);
  compose(bmp, BG_CHAR, SOFT_CHAR, SOFT_CHAR, SOFT_LW, 1);
  compose(bmp, BG_CHAR, MELT_CHAR, MELT_CHAR, MELT_LW, 1);
  compose(bmp, BG_CHAR, NUKE_CHAR, NUKE_CHAR, NUKE_LW, 1);
  compose(bmp, BG_CHAR, ITEM_BG_CHAR, ITEM_BG_CHAR, 1, 1);
  compose(bmp, ITEM_BG_CHAR, ITEM_CHAR, ITEM_CHAR, ITEM_LW, ITEMS);
  compose(bmp, BG_CHAR, LWALL_CHAR, LWALL_CHAR, LWALL_LW, LWALL_H);
  compose(bmp, BG_CHAR, RWALL_CHAR, RWALL_CHAR, RWALL_LW, RWALL_H);
  compose(bmp, 0, HWALL_CHAR, HWALL_CH1, HWALL_LW * HWALL_W, 1);
  compose(bmp, BG_CHAR, HWALL_CHAR, HWALL_CHAR, HWALL_LW * HWALL_W, HWALL_H);
  /* composite time bomb clocks */
  for (y = 1; y < BOMB_ST_TIME_H; y ++)
    for (x = 0; x < BOMB_LW; x ++)
      compose(bmp,
	      CHARSET(BOMB_ST_TIME * BOMB_H, BOMB_CHAR + BOMB_W * x),
	      CHARSET((BOMB_ST_TIME + y) * BOMB_H, BOMB_CHAR + BOMB_W * x),
	      CHARSET((BOMB_ST_TIME + y) * BOMB_H, BOMB_CHAR + BOMB_W * x),
	      BOMB_W, BOMB_H);
  printf("ok\n");
  printf("Initializing: ");
  fflush(stdout);

  /* initialize screen matrices */
  for (y = 0; y < ROWS; y ++) {
     for (x = 0; x < COLS; x ++) {
        if ((x < 3) || (x > 17) || (! y) || (y > 14))
	  blank(y, x);
	else if (x == 3) {
	  vs_screen[0][y][x] . ch = LWALL_CHAR;
	  vs_screen[0][y][x] . lw = LWALL_LW;
	  vs_screen[0][y][x] . delay = LWALL_D;
	  if (y < 3)
	    vs_screen[0][y][x] . st = y - 1;
	  else if (y < 13)
	    vs_screen[0][y][x] . st = LWALL_ST_0 + (y - 3) % LWALL_LH;
	  else
	    vs_screen[0][y][x] . st = LWALL_ST_0 + LWALL_LH + (y - 13);
	  vs_screen[0][y][x] . fr = 0;
	  vs_screen[0][y][x] . timer = vs_screen[0][y][x] . delay;
	  vs_screen[1][y - 1][x] . ch = 0;
	} else if (x == 17) {
	  vs_screen[0][y][x] . ch = RWALL_CHAR;
	  vs_screen[0][y][x] . lw = RWALL_LW;
	  vs_screen[0][y][x] . delay = RWALL_D;
	  if (y < 3)
	    vs_screen[0][y][x] . st = y - 1;
	  else if (y < 13)
	    vs_screen[0][y][x] . st = RWALL_ST_0 + (y - 3) % RWALL_LH;
	  else
	    vs_screen[0][y][x] . st = RWALL_ST_0 + RWALL_LH + (y - 13);
	  vs_screen[0][y][x] . fr = 0;
	  vs_screen[0][y][x] . timer = vs_screen[0][y][x] . delay;
	  vs_screen[1][y - 1][x] . ch = 0;
        } else if (y < 3) {
	  vs_screen[0][y][x] . ch =
	    HWALL_CHAR + (x - 4) % HWALL_W * HWALL_LW;
	  vs_screen[0][y][x] . lw = HWALL_LW;
	  vs_screen[0][y][x] . delay = HWALL_D;
	  vs_screen[0][y][x] . fr = 0;
	  vs_screen[0][y][x] . st = (y - 1) % HWALL_H;
	  vs_screen[0][y][x] . timer = vs_screen[0][y][x] . delay;
	  vs_screen[1][y - 1][x] . ch = 0;
        } else if (y == 14) {
	  vs_screen[0][y][x] . ch =
	    HWALL_CHAR + (x - 4) % HWALL_W * HWALL_LW;
	  vs_screen[0][y][x] . lw = HWALL_LW;
	  vs_screen[0][y][x] . delay = HWALL_D;
	  vs_screen[0][y][x] . fr = 0;
	  vs_screen[0][y][x] . st = (y - 13) % HWALL_H;
	  vs_screen[0][y][x] . timer = vs_screen[0][y][x] . delay;
	  vs_screen[1][y - 1][x] . ch =
	    HWALL_CH1 + (x - 4) % HWALL_W * HWALL_LW;
	  vs_screen[1][y - 1][x] . lw = HWALL_LW;
	  vs_screen[1][y - 1][x] . delay = HWALL_D;
	  vs_screen[1][y - 1][x] . fr = 0;
	  vs_screen[1][y - 1][x] . st = 0;
	  vs_screen[1][y - 1][x] . timer =
	    vs_screen[1][y - 1][x] . delay;
	} else if ((y ^ 1) & x & 1) {
	  vs_screen[0][y][x] . ch = HARD_CHAR;
	  vs_screen[0][y][x] . lw = HARD_LW;
	  vs_screen[0][y][x] . delay = HARD_D;
	  vs_screen[0][y][x] . fr = 0;
	  vs_screen[0][y][x] . st = 0;
	  vs_screen[0][y][x] . timer = vs_screen[0][y][x] . delay;
	  vs_screen[1][y - 1][x] . ch = 0;
        } else if ((((x == 4) || (x == 16)) &&
                    ((y < 6) || (y > 10))) ||
                   (((x < 7) || (x > 13)) &&
                    ((y == 3) || (y == 13))))
	  blank(y, x);
        else {
	  vs_screen[0][y][x] . ch = SOFT_CHAR;
	  vs_screen[0][y][x] . lw = SOFT_LW;
	  vs_screen[0][y][x] . delay = SOFT_D;
	  vs_screen[0][y][x] . fr = 0;
	  vs_screen[0][y][x] . st = 0;
	  vs_screen[0][y][x] . timer = vs_screen[0][y][x] . delay;
	  vs_screen[1][y - 1][x] . ch = 0;
        }
        vs_screen[0][y][x] . dirty = 1;
     }
     putchar(("/-\\|")[y & 3]);
     putchar('\b');
     fflush(stdout);
  }
  /* blow holes in some soft tiles */
  {
    int i, j;

    for (i = 0; i < 12; i ++) {
      do {
        x = rand() % 13 + 3;
        y = rand() % 11 + 2;
	for (j = 0; j < 4; j ++)
	  if (isblank(y + ("\1\2\1\0")[j] - 1,
		      x + ("\2\1\0\1")[j] - 1))
	    break;
      } while ((vs_screen[0][y][x] . ch != SOFT_CHAR) || (j < 4));
      blank(y, x);
      putchar(("/-\\|")[y & 3]);
      putchar('\b');
      fflush(stdout);
    }
  }
  /* initialize sprite priorities */
  for (s = 0; s < SPRITES; s ++) {
     sprite[s] . prio = s;
     sprite_prio[s] = s;
     putchar(("/-\\|")[s & 3]);
     putchar('\b');
     fflush(stdout);
  }
  /* initialize poison marker sprite */
  {
     s = POISON_S;
     sprite[s] . used = 0;
     sprite[s] . z = 0;
     sprite[s] . w = POISON_W;
     sprite[s] . h = POISON_H;
     sprite[s] . x = 0;
     sprite[s] . y = 0;
     sprite[s] . r = 0;
     sprite[s] . c = 0;
     sprite[s] . ch = POISON_CHAR;
     sprite[s] . fr = 0;
     sprite[s] . st = 0;
     sprite[s] . mst = sprite[s] . st;
     sprite[s] . lw = POISON_LW;
     sprite[s] . delay = POISON_D;
     sprite[s] . timer = sprite[s] . delay;
  }
  /* initialize pause sprite */
  {
     s = PAUSE_S;
     sprite[s] . used = 0;
     sprite[s] . z = - screen_h;
     sprite[s] . w = PAUSE_W;
     sprite[s] . h = PAUSE_H;
     sprite[s] . x = (COLS - sprite[s] . w) * char_w / 2;
     sprite[s] . y = - char_h * sprite[s] . h;
     sprite[s] . r = 0;
     sprite[s] . c = 0;
     sprite[s] . ch = PAUSE_CHAR;
     sprite[s] . fr = 0;
     sprite[s] . st = 0;
     sprite[s] . mst = sprite[s] . st;
     sprite[s] . lw = PAUSE_LW;
     sprite[s] . delay = PAUSE_D;
     sprite[s] . timer = sprite[s] . delay;
  }
  /* initialize pause overlay sprite */
  {
     s = OVER_S;
     sprite[s] . used = 0;
     sprite[s] . z = -1 - screen_h;
     sprite[s] . w = OVER_W;
     sprite[s] . h = OVER_H;
     sprite[s] . x = sprite[PAUSE_S] . x;
     sprite[s] . y = sprite[PAUSE_S] . y;
     sprite[s] . r = OVER_R;
     sprite[s] . c = OVER_C;
     sprite[s] . ch = OVER_CHAR;
     sprite[s] . fr = 0;
     sprite[s] . st = 0;
     sprite[s] . mst = sprite[s] . st;
     sprite[s] . lw = OVER_LW;
     sprite[s] . delay = OVER_D;
     sprite[s] . timer = sprite[s] . delay;
  }
  /* initialize bombs and bomb shadows */
  for (b = 0; b < BOMBS; b ++) {
     s = BOMB_S(b);
     sprite[s] . used = 0;
     sprite[s] . x = 0;
     sprite[s] . y = 0;
     sprite[s] . z = 1 - (char_h * sprite[s] . r + sprite[s] . y);
     sprite[s] . w = BOMB_W;
     sprite[s] . h = BOMB_H;
     sprite[s] . ch = BOMB_CHAR;
     sprite[s] . fr = 0;
     sprite[s] . st = BOMB_ST_NORM;
     sprite[s] . mst = sprite[s] . st;
     sprite[s] . lw = BOMB_LW;
     sprite[s] . delay = BOMB_D;
     sprite[s] . timer = sprite[s] . delay;
     s = BSHADOW_S(b);
     sprite[s] . used = 0;
     sprite[s] . x = 0;
     sprite[s] . y = 0;
     sprite[s] . z = sprite[BOMB_S(b)] . z + 1;
     sprite[s] . w = BSHADOW_W;
     sprite[s] . h = BSHADOW_H;
     sprite[s] . ch = BSHADOW_CHAR;
     sprite[s] . fr = 0;
     sprite[s] . st = BOMB_ST_NORM;
     sprite[s] . mst = sprite[s] . st;
     sprite[s] . lw = BSHADOW_LW;
     sprite[s] . delay = BSHADOW_D;
     sprite[s] . timer = sprite[s] . delay;
     putchar(("/-\\|")[s & 3]);
     putchar('\b');
     fflush(stdout);
  }
  /* initialize players */
  for (p = 0; p < PLAYERS; p ++) {
     init_player(p);
     putchar(("/-\\|")[s & 3]);
     putchar('\b');
     fflush(stdout);
  }
  printf("ok\n");
  printf("\nPress ESC to exit, any other key for %dx%dx%d-bit gfx. mode... ",
         screen_w, screen_h, screen_depth);
  fflush(stdout);
  if ((readkey() & 0xff) == 0x1b) {
     printf("exiting!\n");
     return 2;
  }
  putchar('\n');
  if (set_gfx_mode(gfx_card, screen_w, screen_h, 0, 0)) {
     fprintf(stderr,
             "%s: set_gfx_mode failed, reason: %s\n",
             *argv, allegro_error);
     return 1;
  }
  if (screen_depth <= 8)
     set_palette(pal);
  clear(screen);
  update = 0;
  refresh(bmp, unthrottled, upper);
  do {
     /* animate tiles, remove stale explosions */
     for (y = 0; y < ROWS; y ++)
       for (x = 0; x < COLS; x ++)
	 if (vs_screen[0][y][x] . ch &&
	     vs_screen[0][y][x] . timer &&
	     ! -- vs_screen[0][y][x] . timer) {
	   vs_screen[0][y][x] . fr ++;
	   if (! (vs_screen[0][y][x] . fr %= vs_screen[0][y][x] . lw)) {
	     switch (vs_screen[0][y][x] . ch) {
	     case MELT_CHAR:
               if ((! poison) || items) {
		 vs_screen[0][y][x] . ch = ITEM_CHAR;
		 vs_screen[0][y][x] . lw = ITEM_LW;
		 vs_screen[0][y][x] . fr = 0;
                 if (items) {
                    if (poison) {
                      vs_screen[0][y][x] . st = items % (ITEMS - 1);
                      if (vs_screen[0][y][x] . st >= ITEM_ST_POISON)
                        vs_screen[0][y][x] . st ++;
                    } else
                      vs_screen[0][y][x] . st = items % ITEMS;
                 } else
                    vs_screen[0][y][x] . st = ITEM_ST_POISON;
                 if (vs_screen[0][y][x] . st == ITEM_ST_POISON)
                   poison = 1;
		 vs_screen[0][y][x] . delay = ITEM_D;
                 if (items)
                   -- items;
                 break;
               }
	     case NUKE_CHAR:
	     case BOOM_CHAR:
	       blank(y, x);
	       break;
	     }
	   }
	   vs_screen[0][y][x] . timer = vs_screen[0][y][x] . delay;
	   vs_screen[0][y][x] . dirty = 1;
	 }
     /* handle lit bombs (with caution!) */
     for (b = 0; b < BOMBS; b ++) {
        int r, c, x, y, dx, dy, ch, range, s1, st;
       
	s = BOMB_S(b);
        s1 = BSHADOW_S(b);
	if (! sprite[s] . used)
 	  continue;
	r = sprite[s] . r;
	c = sprite[s] . c;
	x = sprite[s] . x;
	y = sprite[s] . y;
	st = sprite[s] . st;
	dx = bomb[b] . dx;
	dy = bomb[b] . dy;
	p = bomb[b] . p;
	ch = vs_screen[0][r][c] . ch;
        if (sprite_loop(s, (int *)0) &&
	    (! sprite[s] . fr) &&
	    (bomb[b] . fuse <= 
	     sprite[s] . delay *
	     2 * (sprite[s1] . lw - 1) *
	     BOMB_ST_TIME_H) &&
	    (st >= BOMB_ST_TIME) &&
	    (st < (BOMB_ST_TIME + BOMB_ST_TIME_H - 1))) {
	  sprite[s] . st = ++ st;
	  dirty_sprite(s);
	}
        sprite_loop(s1, (int *)0);
	if (! key[player[p] . keys[5]])
	  player[p] . special = 0;
	else if (sprite[s] . used &&
	    (! bomb[b] . fuse) &&
	    (st == BOMB_ST_REMOTE) &&
	    ! player[p] . special) {
	  int s2 = MAN_S(p);
	  
	  player[p] . special = 1;
          if (player[p] . alive) {
            bomb[b] . fuse = MAN_D * sprite[s2] . lw;
            sprite[s2] . st = MAN_ST_CLICK;
            sprite[s2] . fr = 0;
            sprite[s2] . delay = MAN_D;
            sprite[s2] . timer = sprite[s2] . delay;
            dirty_sprite(s2);
	  }
	}
        if (sprite[s] . used &&
	    bomb[b] . fuse &&
	    ! (-- bomb[b] . fuse)) {
          sprite[s1] . used = 0;
          dirty_sprite(s1);
	  dirty_sprite(s);
	  sprite[s] . used = 0;
	  dirty_sprite(s);
	  player[p] . active --;
          if (st == BOMB_ST_REMOTE)
	    player[p] . remote ++;
	  else if ((st >= BOMB_ST_TIME) &&
		   (st < (BOMB_ST_TIME + BOMB_ST_TIME_H)))
	    player[p] . time ++;
          if (st == BOMB_ST_SMALL)
            range = 1;
          else 
            range = player[p] . range;
	  /* explosion center */
	  boom_mix(r, c, BOOM_ST_CENTER);
	  /* left arm of explosion */
	  for (x = c - 1;
	       ((vs_screen[0][r][x] . ch == BOOM_CHAR) ||
	        isblank(r, x)) &&
	       (x > c - range);
	       x --)
	    boom_mix(r, x, BOOM_ST_HORIZ);
	  /* left end-cap of explosion */
	  if ((x == c - range) &&
	      ((vs_screen[0][r][x] . ch == BOOM_CHAR) ||
	       isblank(r, x)))
	    boom_mix(r, x, BOOM_ST_LEFT);
	  else
	    blast(r, x);
	  /* right arm of explosion */
	  for (x = c + 1;
	       ((vs_screen[0][r][x] . ch == BOOM_CHAR) ||
		isblank(r, x)) &&
	       (x < c + range);
	       x ++)
	    boom_mix(r, x, BOOM_ST_HORIZ);
	  /* right end-cap of explosion */
	  if ((x == c + range) &&
	      ((vs_screen[0][r][x] . ch == BOOM_CHAR) ||
	       isblank(r, x)))
	    boom_mix(r, x, BOOM_ST_RIGHT);
	  else
	    blast(r, x);
	  /* top arm of explosion */
	  for (y = r - 1;
	       ((vs_screen[0][y][c] . ch == BOOM_CHAR) ||
		isblank(y, c)) &&
	       (y > r - range);
	       y --)
	    boom_mix(y, c, BOOM_ST_VERT);
	  /* top end-cap of explosion */
	  if ((y == r - range) &&
	      ((vs_screen[0][y][c] . ch == BOOM_CHAR) ||
	       isblank(y, c)))
	    boom_mix(y, c, BOOM_ST_TOP);
	  else
	    blast(y, c);
	  /* bottom arm of explosion */
	  for (y = r + 1;
	       ((vs_screen[0][y][c] . ch == BOOM_CHAR) ||
		isblank(y, c)) &&
	       (y < r + range);
	       y ++)
	    boom_mix(y, c, BOOM_ST_VERT);
	  /* bottom end-cap of explosion */
	  if ((y == r + range) &&
	      ((vs_screen[0][y][c] . ch == BOOM_CHAR) ||
	       isblank(y, c)))
	    boom_mix(y, c, BOOM_ST_BOTTOM);
	  else
	    blast(y, c);
	} else if (ch == BOOM_CHAR) {
	  /* chain-reaction explosion */
          if ((bomb[b] . fuse > CHAIN_FUSE) || ! bomb[b] . fuse)
             bomb[b] . fuse = CHAIN_FUSE;
	} else if (dx || dy) {
	  int k;

	  dirty_sprite(s);
	  dirty_sprite(s1);
	  if ((k = kickable(b, r + dy, c + dx)) ||
	      ((dx < 0) && x) ||
	      ((dy < 0) && y)) {
	    x += dx * BOMB_DX;
	    if (x < 0) {
	      if (k) {
		x += char_w;
		c --;
	      } else
		x = 0;
	    } else if (x >= char_w) {
	      x -= char_w;
	      c ++;
	    }
	    y += dy * BOMB_DY;
	    if (y < 0) {
	      if (k) {
		y += char_h;
		r --;
	      } else
		y = 0;
	    } else if (y >= char_h) {
	      y -= char_h;
	      r ++;
	    }
	    sprite[s] . r = r;
	    sprite[s1] . r = r;
	    sprite[s] . c = c;
	    sprite[s1] . c = c;
	    sprite[s] . x = x;
	    sprite[s1] . x = x;
	    sprite[s] . y = y;
	    sprite[s1] . y = y;
	  } else {
	    bomb[b] . dx = 0;
	    bomb[b] . dy = 0;
	    sprite[s] . x = 0;
	    sprite[s1] . x = 0;
	    sprite[s] . y = 0;
	    sprite[s1] . y = 0;
	  }
	  sprite[s] . z = 1 - (char_h * sprite[s] . r + sprite[s] . y);
	  sprite[s1] . z = sprite[s] . z + 1;
	  dirty_sprite(s);
	  dirty_sprite(s1);
	}
     }
     /* control players */
     for (p = 0; p < PLAYERS; p ++) {
        int moved = 0, alive, r, c, ch, dirty = 0, bombs, dx, dy;

	s = MAN_S(p);
        x = sprite[s] . x;
        y = sprite[s] . y;
	r = sprite[s] . r;
	c = sprite[s] . c;
        if (player[p] . poisoned &&
            (player[p] . poison == POISON_SLOW)) {
          dx = MAN_DX_SLOW;
          dy = MAN_DY_SLOW;
        } else if (player[p] . poisoned &&
                   (player[p] . poison == POISON_FAST)) {
          dx = MAN_DX_FAST;
          dy = MAN_DY_FAST;
        } else {
          dx = MAN_DX;
          dy = MAN_DY;
        }
        bombs = player[p] . bombs;
        if (player[p] . poisoned &&
            (player[p] . poison == POISON_SMALL) &&
            bombs)
          bombs = 1;
	sprite[s] . z = - (char_h * (1 + r) + y);
	ch = vs_screen[0][r][c] . ch;
	if (sprite[s] . ch == DEATH_CHAR) {
	   if (sprite_repeat(s, (int *)0)) {
	      sprite[s] . used = 0;
	      dirty = 1;
	   }
	} else if (ch == BOOM_CHAR) {
	  dirty_sprite(s);
	  sprite[s] . ch = DEATH_CHAR;
	  sprite[s] . st = p;
	  sprite[s] . delay = DEATH_D;
	  sprite[s] . timer = sprite[s] . delay;
	  sprite[s] . h = DEATH_H;
	  sprite[s] . w = DEATH_W;
	  sprite[s] . lw = DEATH_LW;
	  sprite[s] . fr = 0;
	  player[p] . alive = 0;
          if (player[p] . poisoned) {
            dirty_sprite(POISON_S);
            sprite[POISON_S] . used = 0;
            dirty_sprite(POISON_S);
            player[p] . poisoned = 0;
            poison = 0;
          }
          dirty = 1;
        } else if (player[p] . alive && (ch == ITEM_CHAR)) {
	  switch (vs_screen[0][r][c] . st) {
	  case ITEM_ST_FLAME:
	     player[p] . range ++;
	     break;
          case ITEM_ST_BOMB:
	     player[p] . bombs ++;
	     break;
          case ITEM_ST_KICK:
	     player[p] . kick = 1;
	     break;
          case ITEM_ST_REMOTE:
             player[p] . remote ++;
	     break;
          case ITEM_ST_POISON:
             player[p] . poisoned = POISON_T;
             player[p] . poison = rand() % POISONS;
             sprite[POISON_S] . used = 1;
             sprite[POISON_S] . x = x;
             sprite[POISON_S] . y = y;
             sprite[POISON_S] . z = sprite[s] . z;
             dirty_sprite(POISON_S);
	     break;
          case ITEM_ST_TIME:
	     player[p] . time ++;
	     break;
          case ITEM_ST_THROW:
	     player[p] . throw = 1;
	     break;
	  }
	  blank(r, c);
	}
	alive = player[p] . alive;
	if (alive &&
	    player[p] . time &&
	    ! player[p] . remote) {
	  if (key[player[p] . keys[4]])
	    player[p] . delay ++;
	}
	if (key[player[p] . keys[5]] &&
	    ! (sprite[s] . used || alive)) {
	  player[p] . special = 1;
	  dirty_sprite(s);
	  init_player(p);
	  r = sprite[s] . r;
	  c = sprite[s] . c;
	  x = sprite[s] . x;
	  y = sprite[s] . y;
	  dirty = 1;
        } else if (alive &&
                   (player[p] . active < bombs) &&
                   ! (player[p] . poisoned &&
                      (player[p] . poison == POISON_BOMB0)) &&
		   ((key[player[p] . keys[4]] &&
		     (player[p] . remote ||
		      ! player[p] . time)) ||
                    (player[p] . poisoned &&
                      (player[p] . poison == POISON_BOMB1)) ||
		    (player[p] . delay &&
		     ! key[player[p] . keys[4]]))) {
          for (b = 0; b < BOMBS; b ++) {
            int s = BOMB_S(b);
	    
            if (sprite[s] . used &&
                (sprite[s] . r == r) &&
                (sprite[s] . c == c)) {
              if (player[p] . throw) {
                /* throw a bomb */
              }
	      break;
            }
          }
          if (b < BOMBS)
            b = BOMBS;
          else
            for (b = 0; b < BOMBS; b ++)
              if (! sprite[BOMB_S(b)] . used)
		break;
	  if ((b < BOMBS) && isblank(r, c)) {
            int s1, s2;
	    
	    player[p] . active ++;
	    bomb[b] . p = p;
	    bomb[b] . dx = 0;
	    bomb[b] . dy = 0;
	    s1 = BOMB_S(b);
            s2 = BSHADOW_S(b);
            if (player[p] . poisoned &&
                (player[p] . poison == POISON_SMALL)) {
              sprite[s1] . st = BOMB_ST_SMALL;
	      bomb[b] . fuse = BOMB_FUSE;
	      sprite[s1] . timer = sprite[s1] . delay;
	      sprite[s2] . timer = sprite[s2] . delay;
            } else if (player[p] . remote) {
	      player[p] . remote --;
	      sprite[s1] . st = BOMB_ST_REMOTE;
	      bomb[b] . fuse = 0;
	      sprite[s1] . timer = sprite[s1] . delay;
	      sprite[s2] . timer = sprite[s2] . delay;
	    } else if (player[p] . time && player[p] . delay) {
	      player[p] . time --;
	      sprite[s1] . st = BOMB_ST_TIME;
	      bomb[b] . fuse = player[p] . delay +
		sprite[s1] . delay *
		  2 * (sprite[s1] . lw - 1) *
		    BOMB_ST_TIME_H;
	      sprite[s1] . timer = sprite[s1] . delay -
		player[p] . delay % sprite[s1] . delay;
	      sprite[s2] . timer = sprite[s2] . delay -
		player[p] . delay % sprite[s2] . delay;
	      player[p] . delay = 0;
	    } else {
	      sprite[s1] . st = BOMB_ST_NORM;
	      bomb[b] . fuse = BOMB_FUSE;
	      sprite[s1] . timer = sprite[s1] . delay;
	      sprite[s2] . timer = sprite[s2] . delay;
	    }
	    sprite[s1] . used = 1;
            sprite[s2] . used = 1;
	    sprite[s1] . r = r;
            sprite[s2] . r = r;
	    sprite[s1] . c = c;
            sprite[s2] . c = c;
	    sprite[s1] . x = 0;
            sprite[s2] . x = 0;
	    sprite[s1] . y = 0;
            sprite[s2] . y = 0;
	    sprite[s1] . z = 1 - (char_h * sprite[s1] . r + sprite[s1] . y);
	    sprite[s2] . z = sprite[s1] . z + 1;
	    sprite[s1] . fr = 0;
            sprite[s2] . fr = 0;
	    dirty_sprite(s1);
            dirty_sprite(s2);
	  }
	}
	if (player[p] . delay && ! key[player[p] . keys[4]])
	  player[p] . delay = 0;
	if (sprite[s] . st == MAN_ST_CLICK)
	  ;
        else if (key[player[p] . keys[0]] && alive) {
           sprite[s] . st = MAN_ST_UP;
           sprite[s] . delay = Y_STEP;
           if (! sprite[s] . timer)
              sprite[s] . timer = sprite[s] . delay;
           if ((y > -char_h) ||
               (walkable(p, 0, -1))) {
              dirty_sprite(s);
              moved = 1;
              sprite[s] . mst = MAN_ST_UP;
              if ((y -= dy) < (3 * char_h) / -2) {
                 y += char_h;
                 sprite[s] . r --;
              }
              if (x < 0) {
                if ((x += dx) > 0)
		  x = 0;
	      } else if (x > 0) {
                if ((x -= dx) < 0)
		  x = 0;
	      }
           } else {
	     if (y != -char_h) {
	       dirty_sprite(s);
	       y = -char_h;
	     }
	     sprite[s] . delay *= 2;
	   }
           dirty = 1;
        } else if (key[player[p] . keys[1]] && alive) {
           sprite[s] . st = MAN_ST_DOWN;
           sprite[s] . delay = Y_STEP;
           if (! sprite[s] . timer)
              sprite[s] . timer = sprite[s] . delay;
           if ((y < -char_h) ||
               (walkable(p, 0, 1))) {
              dirty_sprite(s);
              moved = 1;
              sprite[s] . mst = MAN_ST_DOWN;
              if ((y += dy) > char_h / -2) {
                 y -= char_h;
                 sprite[s] . r ++;
              }
              if (x < 0) {
                if ((x += dx) > 0)
		  x = 0;
	      } else if (x > 0) {
                if ((x -= dx) < 0)
		  x = 0;
	      }
           } else {
	     if (y != -char_h) {
	       dirty_sprite(s);
	       y = -char_h;
	     }
	     sprite[s] . delay *= 2;
	   }
           dirty = 1;
        }
	if (moved || (sprite[s] . st == MAN_ST_CLICK))
	  ;
        else if (key[player[p] . keys[2]] && alive) {
           if ((! dirty) || (sprite[s] . mst == MAN_ST_LEFT))
              sprite[s] . st = MAN_ST_LEFT;
           sprite[s] . delay = X_STEP;
           if (! sprite[s] . timer)
              sprite[s] . timer = sprite[s] . delay;
           if ((x > 0) ||
               (walkable(p, -1, 0))) {
              dirty_sprite(s);
              moved = 1;
              sprite[s] . mst = MAN_ST_LEFT;
              if ((x -= dx) < char_w / -2) {
                 x += char_w;
                 sprite[s] . c --;
              }
              if (y < -char_h) {
                if ((y += dy) > -char_h)
		  y = -char_h;
	      } else if (y > -char_h) {
                if ((y -= dy) < -char_h)
		  y = -char_h;
	      }
           } else {
	     if (x) {
	       dirty_sprite(s);
	       x = 0;
	     }
	     sprite[s] . delay *= 2;
	   }
           dirty = 1;
        } else if (key[player[p] . keys[3]] && alive) {
           if ((! dirty) || (sprite[s] . mst == MAN_ST_RIGHT))
              sprite[s] . st = MAN_ST_RIGHT;
           sprite[s] . delay = X_STEP;
           if (! sprite[s] . timer)
              sprite[s] . timer = sprite[s] . delay;
           if ((x < 0) ||
               (walkable(p, 1, 0))) {
              dirty_sprite(s);
              moved = 1;
              sprite[s] . mst = MAN_ST_RIGHT;
              if ((x += dx) > char_w / 2) {
                 x -= char_w;
                 sprite[s] . c ++;
              }
              if (y < -char_h) {
                if ((y += dy) > -char_h)
		  y = -char_h;
	      } else if (y > -char_h) {
                if ((y -= dy) < -char_h)
		  y = -char_h;
	      }
           } else {
	     if (x) {
	       dirty_sprite(s);
	       x = 0;
	     }
	     sprite[s] . delay *= 2;
	   }
           dirty = 1;
	}
        if (moved) {
           sprite[s] . st = sprite[s] . mst;
	   sprite[s] . z = - (char_h * (1 + r) + y);
	}
        if (alive &&
	    (! dirty) &&
	    (sprite[s] . fr != 1) &&
	    (sprite[s] . st != MAN_ST_CLICK)) {
           sprite[s] . timer = 0;
           sprite[s] . fr = 1;
           dirty = 1;
        }
        sprite[s] . x = x;
        sprite[s] . y = y;
        if (sprite_loop(s, &dirty) &&
	    alive &&
	    (sprite[s] . st == MAN_ST_CLICK) &&
	    ! sprite[s] . fr) {
	  sprite[s] . timer = 0;
	  sprite[s] . st = MAN_ST_DOWN;
	  sprite[s] . fr = 1;
	  dirty = 1;
	}
        if (dirty)
           dirty_sprite(s);
        if (player[p] . poisoned && player[p] . alive) {
          int vfr = 0;

          sprite_loop(POISON_S, (int *)0);
          dirty_sprite(POISON_S);
          if (-- player[p] . poisoned) {
            int p1, s1;

            for (p1 = 0; p1 < PLAYERS; p1 ++) {
               s1 = MAN_S(p1);
               if ((p1 != p) &&
                   player[p1] . alive &&
                   (sprite[s1] . r == sprite[s] . r) &&
                   (sprite[s1] . c == sprite[s] . c) &&
                   (! player[p1] . poisoned) &&
                   ! player[p1] . vector) {
                 player[p] . vector = 1;
                 player[p1] . poisoned = POISON_T;
                 player[p1] . poison = player[p] . poison;
                 player[p] . poisoned = 0;
                 break;
               }
            }
            if (! (p1 < PLAYERS))
               s1 = s;
            vfr = vframe % char_h;
            if (vfr > (char_h / 2) + 1)
              vfr = char_h - 1 - vfr;
            sprite[POISON_S] . y = sprite[s1] . y - char_h + vfr;
            sprite[POISON_S] . r = sprite[s1] . r;
            sprite[POISON_S] . c = sprite[s1] . c;
            sprite[POISON_S] . x = sprite[s1] . x;
            sprite[POISON_S] . z = sprite[s1] . z;
          } else {
            poison = 0;
            sprite[POISON_S] . used = 0;
          }
          dirty_sprite(POISON_S);
        } else if (player[p] . vector) {
          int p1, s1;

          for (p1 = 0; p1 < PLAYERS; p1 ++) {
             s1 = MAN_S(p1);
             if ((p1 != p) &&
                 player[p1] . alive &&
                 (sprite[s1] . r == sprite[s] . r) &&
                 (sprite[s1] . c == sprite[s] . c))
               break;
          }
          if (! (p1 < PLAYERS))
            player[p] . vector = 0;
        }
     }
     /* pause */
     if (key[KEY_F1] || key[KEY_P]) {
       while ((key[KEY_F1] || key[KEY_P]) && ! key[KEY_ESC]);
       sprite[PAUSE_S] . used = 1;
       for (y = - sprite[PAUSE_S] . h * char_h;
            y < ((ROWS - sprite[PAUSE_S] . h) * char_h / 2);
            y += PAUSE_DY) {
         dirty_sprite(PAUSE_S);
         sprite[PAUSE_S] . y = y;
         dirty_sprite(PAUSE_S);
	 refresh(bmp, unthrottled, 0);
       }
       while ((key[KEY_F1] || key[KEY_P]) && ! key[KEY_ESC]);
       sprite[OVER_S] . used = 1;
       sprite[OVER_S] . timer = BLINK_D;
       sprite[OVER_S] . fr = 0;
       sprite[OVER_S] . y = sprite[PAUSE_S] . y;
       dirty_sprite(OVER_S);
       while (! ((key[KEY_F1] || key[KEY_P]) || key[KEY_ESC])) {
	 refresh(bmp, unthrottled, 0);
	 sprite_loop(PAUSE_S, (int *)0);
	 if (sprite_loop(OVER_S, (int *)0))
	   sprite[OVER_S] . timer = BLINK_D;
       }
       while ((key[KEY_F1] || key[KEY_P]) && ! key[KEY_ESC]);
       dirty_sprite(OVER_S);
       sprite[OVER_S] . used = 0;
       while (y > - sprite[PAUSE_S] . h * char_h) {
         dirty_sprite(PAUSE_S);
         sprite[PAUSE_S] . y = (y -= PAUSE_DY);
         dirty_sprite(PAUSE_S);
	 refresh(bmp, unthrottled, 0);
       }
       dirty_sprite(PAUSE_S);
       sprite[PAUSE_S] . used = 0;
       update = 0;
       redraw = 1;
     }
     /* toggle renderers */
     if (key[KEY_F9]) {
       tse_style ++;
       tse_style %= TSE_STYLES;
       while (key[KEY_F9]);
       redraw = 1;
     }
     /* toggle upper layer */
     if (key[KEY_F10]) {
       upper = 1 - upper;
       while (key[KEY_F10]);
       redraw = 1;
     }
     /* redraw */
     if (key[KEY_F11])
       redraw = 1;
     if (redraw) {
       redraw = 0;
       for (y = 0; y < ROWS; y ++)
	 for (x = 0; x < COLS; x ++)
	   vs_screen[0][y][x] . dirty = 1;
     }
     /* shot */
     if (key[KEY_F12]) {
        BITMAP *visible;
        int w, h;

        if (x_offset >= 0) {
           x = x_offset;
           w = COLS * char_w;
        } else {
           x = 0;
           w = screen_w;
        }
        if (y_offset >= 0) {
           y = y_offset;
           h = ROWS * char_h;
        } else {
           y = 0;
           h = screen_h;
        }
        if ((visible = create_sub_bitmap(screen, x, y, w, h))) {
           save_bitmap(shot, visible, pal);
           destroy_bitmap(visible);
        }
        while (key[KEY_F12]);
     }
     refresh(bmp, unthrottled, upper);
  } while (! key[KEY_ESC]);
  fade_out(4);
  allegro_exit();
  printf("%s: rendered %d frames (of %d virtual frames)\n" \
	 "\n" \
	 "[vs]\n" \
	 "frameskip = %d\n" \
	 "gfx_card = %d\n" \
	 "gfx_res = %dx%dx%d\n" \
	 "tse_style = %d\n" \
	 "char_size = %dx%d\n" \
	 "speed = %d\n" \
	 "tiles = %s\n",

         *argv, frame, vframe,
         frameskip,
	 gfx_card,
	 screen_w, screen_h, screen_depth,
	 tse_style,
	 char_w, char_h,
	 speed,
	 tiles);
  printf("\nIf you like VS Bomber, you'll love Lung Ripper Champion!\n");
  return 0;
}
